*************************************************************************************
*														*
*	Rain impact simulation V2 for the EASy68k simulator	2006/11/15			*
*														*
*	This code uses double buffering to render a patterm similar to that made	*
*	by raindrops landing in shallow water. It's all a big fudge and mind		*
*	numbingly slow on anything less than a 1.5GHz PC but makes some pretty		*
*	patterns and shows how such effects can be made.					*
*														*
*	This version is faster than the original version but the code is harder		*
*	to follow and greater liberties have been taken with ignoring some error	*
*	conditions in handling the data, this is noticeable but does not detract	*
*	from the overall effect. Also, the background texture can now be toggled	*
*	on and off using the [B] key and four available palettes can be cycled		*
*	through using the [P] key.									*
*														*
*	This version for Sim68K 3.2 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://www.themotionstore.com/leeedavison/index.html					*
*														*
*	.. or ..												*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************


size	EQU	$40					* both x and y size of the pixel array
size2	EQU	size*size-1				* maximum array index, array is indexed 0 to n
mask	EQU	$03					* mask for new drop generation. chance is
							* 1-(n/(n+1)). for lighter rain make this $07

	ORG	$1000

start
	BSR		Initialise			* go setup stuff, pointers, arrays and flags

	MOVE.b	#17,d1			* enable double buffering
	MOVE.b	#92,d0			* set draw mode
	TRAP		#15

	MOVE.b	#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen
	TRAP		#15

* the main loop, after the setup is complete everything is called from within this code

loop
	MOVE.b	#94,d0			* copy screen buffer to main
	TRAP		#15

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	MOVE.l	d1,-(sp)			* push time on the stack, erk! needed d7

	MOVE.b	#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen buffer
	TRAP		#15

	BSR.s		c_controls			* go check the [B] and [P] keys
	BSR.s		gen_array			* generate and draw the new screen data

* now wait for a while. if you have a slow machine this shouldn't slow it down much but
* if you have a fast machine them this will stop the effect running too fast

	MOVE.l	(sp)+,d7			* restore time in 1/100 ths from stack
wait_100ms
	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	SUB.l		d7,d1				* subtract previous time from current time
	CMP.b		#$09,d1			* compare with 9/100ths
	BMI.s		wait_100ms			* loop if time not up yet

* if you want to see how fast it can go comment out the BMI.s above. on my 1.8GHz Celeron
* it does each frame in about 140 to 180ms

	BRA.s		loop				* loop forever

*************************************************************************************
*
* check the [B] and [P] keys. if [B] has just been pressed then toggle the viewmode
* between textured or untextured background view. if [P] has just been pressed step
* the palette 'page' on to the next one

c_controls
	MOVE.l	#'PPBB',d1			* [P]alette and [B]ackground keys
	MOVE.b	#19,d0			* check for keypress
	TRAP		#15

	MOVE.w	d1,d0				* copy result, $0000 if [B] key up, $FFFF if
							* [B] key down
	EOR.w		d1,lastkey(a3)		* compare with last B key state
	BEQ.s		notogglemode		* if state not changed skip toggle

	AND.w		d0,d1				* make sure just pressed, $FFFF if pressed,
							* $0000 if released
	AND.b		#$FE,d1			* mask for 2 x 2 pixel tiles
	EOR.b		d1,viewmode(a3)		* toggle view mode b0 if pressed
notogglemode
	MOVE.w	d0,lastkey(a3)		* save this result as last

	SWAP		d1				* get [P]alette key result

	MOVE.w	d1,d0				* copy result, $0000 if [P] key up, $FFFF if
							* [P] key down
	EOR.w		d1,lastpal(a3)		* compare with last P key state
	BEQ.s		notogglepal			* if state not changed skip toggle

	AND.w		d0,d1				* make sure just pressed, $FFFF if pressed,
							* $0000 if released
	AND.b		#$40,d1			* mask for palette page offset
	ADD.b		d1,palette(a3)		* toggle palette page if pressed
notogglepal
	MOVE.w	d0,lastpal(a3)		* save this result as last
	RTS

*************************************************************************************
*
* generate the next stage and display it

gen_array
	BSR		LAB_RND			* call the PRNG code
	MOVEQ		#mask,d1			* set rain mask
	AND.b		PRNlword(a3),d1		* mask byte from the result
	BNE.s		no_raindrop			* if not time for another drop skip creating one

	MOVE.w	PRNlword+2(a3),d1		* get high word from PRNG result, select any
							* array cell
	AND.w		#size2,d1			* mask to array bounds
	MOVE.b	#$FF,(a1,d1)		* set new raindrop

* for each pixel the new pixel value = (sum of the eight surrounding pixels)/4 - previous
* value. this is a big fudge, any more and it feeds back and just makes noise, any less
* and the drops have too short a life. there is no decay value for the ripples, that is
* taken care of by the loss to rounding on each step

no_raindrop
	MOVEQ		#size-1,d7			* set initial y index
	MOVEQ		#$1C,d2			* set initial screen y1 co-ordinate
	MOVEQ		#$1F,d4			* set initial screen y2 co-ordinate, make a
							* small square

	MOVEQ		#size-1,d5			* set initial x-1 index, time is saved by
							* moving the initial setting of this outside
							* both loops. within the loops it stays in range
							* by being masked at each itteration

	LEA		size*size(a2),a0		* initial y base address, array 2
	LEA		size*size(a1),a5		* initial y base address, array 1
	MOVEA.l	a1,a6				* initial y + 1 base address, array 1
	dc.w		$3C3C				* skip y + 1 address copy for initial element.
							* this makes the next line MOVE.w #$2C4D,d6
							* which is faster than branching round it

* lots of time is saved by making use of the registers. for each y loop the next y-1, y
* and y+1 base addresses are calculated for array 1 as well as the next y base address
* for array 2

gen_yloop
	MOVEA.l	a5,a6				* next y + 1 base address, array 1. this line
							* isn't executed the first time through
	LEA		-size(a0),a0		* next y base address, array 2
	LEA		-size(a5),a5		* next y base address, array 1
	LEA		-size(a5),a4		* next y-1 base address, array 1
	TST.w		d7				* test d7
	BNE.s		gen_yloop_e			* skip address reset if not zeroeth element

	LEA		size*size(a4),a4		* else set y - 1 to top

* calculate the new screen y co-ordinate for this row of pixels. time is saved by having
* this outside the x loop

gen_yloop_e
	ADDQ.w	#4,d2				* calculate the new screen y1 co-ordinate
	ADDQ.w	#4,d4				* calculate the new screen y2 co-ordinate

* set the screen x2 co-ordinate for the first pixel on this row, this saves time as the
* calculation for the next screen x co-ordinates reduces to a copy, an add and a subtract

	MOVEQ		#size-1,d6			* set initial x index
	MOVEQ		#$1F,d3			* set initial screen x2 co-ordinate

* we have y-1, y and y+1 in a4, a5 and a6, now for each x loop first calculate the x-1
* and x+1 values and use these as offsets to the y base addresses to access the values
* of the eight adjoining pixels to the current pixel

* the x loop is repeated once for every pixel so any cycles that can be saved from inside
* this loop will save size*size cycles per frame which for a 64 x 64 array is 4096 cycles

gen_xloop
	MOVEQ		#size-1,d0			* set mask and clear the three highest bytes

	SUBQ.w	#1,d5				* decrement x - 1
	AND.w		d0,d5				* mask to array size

	MOVE.w	d6,d1				* set x index
	ADDQ.w	#1,d1				* add 1
	AND.w		d0,d1				* mask to array size

* here we have x-1, x and x+1 in d5, d6 and d1. using these with a4, a5 and a6 allows easy
* access to the pixel elements

* now get the total for the eight pixels surrounding x,y. it is possible that the running
* total for all eight will be exceede a byte, mainly when a new raindrop is spawned inside
* the leading edge of a previous drop, but this is traded off against the increase in
* speed. errors caused by ignoring this carry are seen in the attenuation of some newly
* spawned drops

	MOVE.b	(a4,d5.w),d0		* get pixel intensity (-1, -1)
	ADD.b		(a4,d6.w),d0		* add pixel intensity ( 0, -1)
	ADD.b		(a4,d1.w),d0		* add pixel intensity (+1, -1)

	ADD.b		(a5,d5.w),d0		* add pixel intensity (-1,  0)
	ADD.b		(a5,d1.w),d0		* add pixel intensity (+1,  0)

	ADD.b		(a6,d5.w),d0		* add pixel intensity (-1, +1)
	ADD.b		(a6,d6.w),d0		* add pixel intensity ( 0, +1)
	ADD.b		(a6,d1.w),d0		* add pixel intensity (+1, +1)

	BEQ.s		pos_pixel			* if the total value is zero just go save and
							* display, this saves some time overall

	LSR.b		#2,d0				* make surrounding pixel total\4, don't use ASR
							* as this is an unsigned byte value
	SUB.b		(a0,d6.w),d0		* subtract pixel from surrounding pixel total\4
	BCC.s		pos_pixel			* if value +ve go save and display

	MOVEQ		#0,d0				* else use zero, don't go blacker than black
pos_pixel
	MOVE.b	d0,(a0,d6.w)		* save new pixel intensity

* the following four lines add a formula texture to the background based on the x,y
* co-ordinates of each pixel and whether the viewmode flag is set

	MOVE.b	d7,d1				* get y co-ordinate
	EOR.b		d6,d1				* combine with x co-ordinate
	AND.b		viewmode(a3),d1		* mask with the view mode flag
	EOR.b		d1,d0				* modify the intensity with the x,y texture

* this selects between the four palette effect 'pages'. OR is used so the total of page
* and intensity never overflows the byte value

	OR.b		palette(a3),d0		* OR with palette flag, select page

* now set the intensity for this possibly modified pixel as the draw colour and the fill
* colour

	ADD.w		d0,d0				* * 2, high byte cleared by previous MOVEQ d0
	ADD.w		d0,d0				* * 4 bytes per longword colour value
	MOVE.l	(a3,d0.w),d1		* get the longword colour value from the table

	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVEQ		#81,d0			* set fill colour
	TRAP		#15

* calculate the screen x co-ordinate for this pixel

	ADDQ.w	#4,d3				* calculate the new screen x2 co-ordinate
	MOVE.w	d3,d1				* copy to the screen x1 co-ordinate
	SUBQ.w	#3,d1				* make a small square

* now draw a small 4 x 4 square of pixels

	MOVEQ		#87,d0			* draw filled rectangle
	TRAP		#15				* do draw

	DBF		d6,gen_xloop		* decrement x and loop for next

	DBF		d7,gen_yloop		* decrement y and loop for next

	EXG		a1,a2				* swap the current and last arrays
	RTS

*************************************************************************************
*
* setup stuff. set the pointers and view mode, initialise the PRNG seed, clear the
* palette page and key flag, clear the background key flag and clear the arrays

Initialise
	LEA		(LAB_VARS,PC),a3		* get pointer to variables
	LEA		array1(a3),a1		* get pointer to array 1
	LEA		array2(a3),a2		* get pointer to array 2
	MOVE.b	#$FE,viewmode(a3)		* set view mode to textured background
	MOVE.l	#$31415927,PRNlword(a3)	* set PRNG seed. this can be any value
							* except zero, changing this will produce
							* a different start point for the effect
	MOVEQ		#0,d0				* clear d0
	MOVE.w	d0,palette(a3)		* clear the palette page
	MOVE.w	d0,lastpal(a3)		* clear the last P key flag
	MOVE.w	d0,lastkey(a3)		* clear the last B key flag
	MOVE.l	#size*size/2-1,d1		* get array size/2
	MOVEA.l	a1,a4				* copy array memory start
clear_loop
	MOVE.l	d0,(a4)+			* clear longword
	DBF		d1,clear_loop		* decrement count and loop for next

	RTS

*************************************************************************************
*
* This is the code that generates the pseudo random sequence. A seed word located in
* PRNlword(a3) is loaded into a register before being operated on to generate the
* next number in the sequence. This number is then saved as the seed for the next
* time it's called.
*
* This code is adapted from the 32 bit version of RND(n) used in EhBASIC68. Taking
* the 19th next number is slower but helps to hide the shift and add nature of this
* generator as can be seen from analysing the output.

LAB_RND
	MOVEM.l	d0-d2,-(sp)			* save d0, d1 and d2
	MOVE.l	PRNlword(a3),d0		* get current seed longword
	MOVEQ		#$AF-$100,d1		* set EOR value
	MOVEQ		#18,d2			* do this 19 times
Ninc0
	ADD.l		d0,d0				* shift left 1 bit
	BCC.s		Ninc1				* if bit not set skip feedback

	EOR.b		d1,d0				* do Galois LFSR feedback
Ninc1
	DBF		d2,Ninc0			* loop

	MOVE.l	d0,PRNlword(a3)		* save back to seed longword
	MOVEM.l	(sp)+,d0-d2			* restore d0, d1 and d2

	RTS

*************************************************************************************
*
* variables used

LAB_VARS						* points to variables base

* colour table for converting byte colour values into longword colour values, this saves
* 16 cycles for each conversion or 65536 cycles per 64 * 64 frame

* another advantage of using a table is that a colour map may be applied that transforms
* the linear grey scale to non linear or even to a coloured scale for such things as
* flame or inverse effects. colours are $00bbggrr

* normal palette
	dc.l	$000000,$010101,$020202,$030303,$040404,$050505,$060606,$070707
	dc.l	$080808,$090909,$0A0A0A,$0B0B0B,$0C0C0C,$0D0D0D,$0E0E0E,$0F0F0F
	dc.l	$101010,$111111,$121212,$131313,$141414,$151515,$161616,$171717
	dc.l	$181818,$191919,$1A1A1A,$1B1B1B,$1C1C1C,$1D1D1D,$1E1E1E,$1F1F1F
	dc.l	$202020,$212121,$222222,$232323,$242424,$252525,$262626,$272727
	dc.l	$282828,$292929,$2A2A2A,$2B2B2B,$2C2C2C,$2D2D2D,$2E2E2E,$2F2F2F
	dc.l	$303030,$313131,$323232,$333333,$343434,$353535,$363636,$373737
	dc.l	$383838,$393939,$3A3A3A,$3B3B3B,$3C3C3C,$3D3D3D,$3E3E3E,$3F3F3F

* 4 x scale palette
	dc.l	$000000,$040404,$080808,$0C0C0C,$101010,$141414,$181818,$1C1C1C
	dc.l	$202020,$242424,$282828,$2C2C2C,$303030,$343434,$383838,$3C3C3C
	dc.l	$404040,$444444,$484848,$4C4C4C,$505050,$545454,$585858,$5C5C5C
	dc.l	$606060,$646464,$686868,$6C6C6C,$707070,$747474,$787878,$7C7C7C
	dc.l	$808080,$848484,$888888,$8C8C8C,$909090,$949494,$989898,$9C9C9C
	dc.l	$A0A0A0,$A4A4A4,$A8A8A8,$ACACAC,$B0B0B0,$B4B4B4,$B8B8B8,$BCBCBC
	dc.l	$C0C0C0,$C4C4C4,$C8C8C8,$CCCCCC,$D0D0D0,$D4D4D4,$D8D8D8,$DCDCDC
	dc.l	$E0E0E0,$E4E4E4,$E8E8E8,$ECECEC,$F0F0F0,$F4F4F4,$F8F8F8,$FFFFFF

* inverse palette
	dc.l	$3F3F3F,$3E3E3E,$3D3D3D,$3C3C3C,$3B3B3B,$3A3A3A,$393939,$383838
	dc.l	$373737,$363636,$353535,$343434,$333333,$323232,$313131,$303030
	dc.l	$2F2F2F,$2E2E2E,$2D2D2D,$2C2C2C,$2B2B2B,$2A2A2A,$292929,$282828
	dc.l	$272727,$262626,$252525,$242424,$232323,$222222,$212121,$202020
	dc.l	$1F1F1F,$1E1E1E,$1D1D1D,$1C1C1C,$1B1B1B,$1A1A1A,$191919,$181818
	dc.l	$171717,$161616,$151515,$141414,$131313,$121212,$111111,$101010
	dc.l	$0F0F0F,$0E0E0E,$0D0D0D,$0C0C0C,$0B0B0B,$0A0A0A,$090909,$080808
	dc.l	$070707,$060606,$050505,$040404,$030303,$020202,$010101,$000000

* flame palette
	dc.l	$000000,$000005,$00000B,$000012,$00001A,$000022,$00002B,$000034
	dc.l	$00003F,$000049,$000054,$00005F,$00006A,$000076,$000082,$00008D
	dc.l	$000098,$0000A4,$0000AE,$0000B9,$0000C3,$0000CD,$0000D7,$0000DF
	dc.l	$0000E7,$0000EF,$0000F5,$0000FB,$0002FD,$0007FD,$000EFD,$0015FD
	dc.l	$001DFD,$0026FD,$0030FD,$003AFD,$0044FD,$0050FD,$005AFD,$0066FD
	dc.l	$0071FD,$007DFD,$0088FD,$0093FD,$009EFD,$00A8FD,$00B2FD,$00BBFD
	dc.l	$00C4FD,$00CBFD,$00D2FD,$00D7FD,$00DCFD,$00DFFD,$00E2FD,$03E0FD
	dc.l	$34E4FD,$66E8FD,$7CECFD,$96F0FD,$BEF4FE,$CAF8FE,$E2FCFE,$FFFFFF
colourend

	OFFSET	colourend-LAB_VARS	* use relative addressing

PRNlword	ds.l	1				* PRNG seed long word

palette	ds.w	1				* palette page
lastpal	ds.w	1				* last P key flag

viewmode	ds.w	1				* view mode
lastkey	ds.w	1				* last B key flag

array1	ds.b	size*size			* make space for the first array
array2	ds.b	size*size			* make space for the second array
arrayend

	ORG	LAB_VARS+arrayend			* use program counter addressing

	END	start
